      program cals3k2coefs

c     A program to calculate Schmidt normalised SHA coefficients for a particular 
c     time from time dependent coefficients in model file CALS3K.2.
c     Output of main field and secular variation coefficients into files 
c     coefs.dat and svcoefs.dat.
c     Output file format: 4 columns with 
c     spherical harmonic degree l, order m, g_l^m, h_l^m 
c
c     Monika Korte, November 2002, version for CALS3K.1
c                   June 2004
c
c     Uses pieces of code from the example program for evaluating 
c     time-dependent field model GUFM
c
c     by Jeremy Bloxham & Andrew Jackson,
c
c     which uses code by: David Gubbins, Kathy Whaler, David Barraclough,
c                         Rick O'Connell, and Carl de Boor
c
c---------------------------------------------------------------------      
c     CALLS:    interv   - calculates which knot lies immediately left
c                          of the current time point
c
c               bspline  - calculates B-splines at current time point
c
c               bspline1 - calculates first time derivative of 
c                          B-splines at current time point
c---------------------------------------------------------------------

      implicit none

      integer lmax,n,nsplt,jord
      parameter (lmax=10)
      parameter (nsplt=55)
      parameter (n=lmax*(lmax+2))

      real*8 g(n),gd(n),gt(n,nsplt),h0
      real*8 time,tknts(nsplt+4),spl(nsplt)
      real*8 tstartin,tendin
      integer lm,nm,i,k,j,l,m,nspl,nleft,itime
      character*50 outfile,outfile2

      data jord/4/

      write(*,*) '                -- CALS3K2coefs --' 
      write(*,*) 'determies Gauss coefficients for an input epoch'
      write(*,*) 'from time-dependent coefficients in file CALS3K.2'
      write(*,*)
      write(*,*) 'Give input year (integer):'
      read(*,*) itime

      time=dble(itime)
      outfile='coefs.dat'
      outfile2='svcoefs.dat'

c     read model
      open(7,file='CALS3K.2')

      read(7,*) tstartin,tendin
      if (tstartin .gt. time .or. tendin .lt. time) then
         write(*,*) 'Error: Input time out of range of model.'
        stop 'Enter a time between 1000 BC and 1950 AD!'
      endif
      read(7,*) lm,nm,nspl,(tknts(i),i=1,nspl+4)
      if(lm.lt.lmax) then
        stop 'lmax inconsistent'
      endif
      read(7,*) gt
      close(7)


c     calculate main field coefficients at time time
c
      
10    call interv(tknts,time,nspl,nleft)
      call bspline(tknts,time,nspl,jord,nleft,spl(nleft-3))
      
      do  k=1,n
       g(k)=0.0
       do j=1,4
        g(k) = g(k) + spl(j+nleft-4)*gt(k,j+nleft-4)
       enddo 
      enddo 
          
c
c     calculate secular variation coefficients at time time

      call bspline1(tknts,time,nspl,nleft,spl)      

      do  k=1,n
       gd(k)=0.0
       do j=1,4
        gd(k) = gd(k) + spl(j+nleft-4)*gt(k,j+nleft-4)
       enddo 
      enddo 

c     output coeffients and SV coefficients:
      open(7,file=outfile)
      open(8,file=outfile2)
      h0=0.0
 5000 format(2I3,2F20.10)
 5001 format(2I3,2F20.10)
      do l=1,lmax
         m=0
         k=l*l
         write(7,*) l,m,g(k),h0
         write(8,*) l,m,gd(k),h0
         do m=1,l
            k=l*l+2*m-1
            write(7,*) l,m,g(k),g(k+1)
            write(8,*) l,m,gd(k),gd(k+1)
         end do
      end do
      close(7)
      close(8)

      end

c     -----------------------------

      subroutine interv(tknts,time,nspl,nleft)
      implicit real*8(a-h,o-z)
      dimension tknts(nspl+1)
      
      if(time.lt.tknts(4).or.time.gt.tknts(nspl+1)) return
      
      do 200 n=5,nspl+1
       if(time.le.tknts(n)) then
        nleft=n-1
        goto 210
       endif
200   continue
210   continue

      return
      end

c --------------------------------------------------

       subroutine bspline(tknts,t,nspl,jorder,nleft,spl)
       implicit real*8(a-h,o-z)
c calculate splines of order jorder where 1 <= jorder <= 4
       
       dimension tknts(nspl+4)
       dimension spl(4)
       
       dimension deltal(4),deltar(4)
       
       spl(1)=1.0
      
       do 200 j=1,jorder-1
       
       deltar(j) = tknts(nleft+j) - t
       deltal(j) = t - tknts(nleft+1-j)
       saved=0.0
       
       do 100 i=1,j
        term = spl(i)/(deltar(i)+deltal(j+1-i))
        spl(i) = saved + deltar(i)*term
        saved = deltal(j+1-i)*term
100    continue

       spl(j+1) = saved
       
200    continue
 
       return
       end

c     --------------------------------------

      subroutine bspline1(tknts,t,nspl,nl,spl)
      implicit real*8(a-h,o-z)
      parameter(nsplmx=164)
      dimension tknts(nspl+4)
      dimension spl(nspl),spl1(nsplmx)
     
      data jord/3/
      
      if(nspl.gt.nsplmx)then
      write(6,*)' increase dimensions of spl1 in bspline1'
      stop
      endif
      
      do 100 is=1,nspl
       spl(is)=0.0
100   continue
      
      call bspline(tknts,t,nspl,jord,nl,spl1(nl-2))
      
      spl(nl) = b0(tknts,nl)*spl1(nl)
      
      spl(nl-1) = b0(tknts,nl-1)*spl1(nl-1) + b1(tknts,nl-1)*spl1(nl)
      spl(nl-2) = b0(tknts,nl-2)*spl1(nl-2) + b1(tknts,nl-2)*spl1(nl-1)
      spl(nl-3) = b1(tknts,nl-3)*spl1(nl-2)
      
      return
      end

      function b0(tknts,i)

      dimension tknts(*)
      implicit real*8(a-h,o-z)      
      b0 = 3.0/( tknts(i+3) - tknts(i) )
 
      return
      end      
      
      function b1(tknts,i)
      implicit real*8(a-h,o-z)
      dimension tknts(*)
      
      b1 = -3.0/( tknts(i+4) - tknts(i+1) )
 
      return
      end
      
